# Copyright 2022-present Lenovo
# Confidential and Proprietary

import argparse
import csv
import os
import json
import time
from itertools import chain

import cv2
import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.driver as cuda
import tensorrt as trt

TRT_LOGGER = trt.Logger()


def get_image_size(model_name):
    if not model_name.find('lenet'):
        image_size = 28
    elif not model_name.find('cifarnet'):
        image_size = 32
    elif not model_name.find('inception_v3'):
        image_size = 299
    elif not model_name.find('inception_v4'):
        image_size = 299
    elif not model_name.find('inception_resnet_v2'):
        image_size = 299
    else:
        image_size = 224

    return image_size


def per_image_standardization(img):
    mean = np.mean(img)
    std = np.std(img)
    count = 1
    for size in img.shape:
        count = count * size
    adj_std = max(std, 1.0 / np.sqrt(count))
    img = (img - mean) / adj_std
    return img


def load_image(image_file, image_size):
    np_img = cv2.imread(image_file, flags=1)
    np_img = np_img.astype(np.float32)
    np_img = cv2.resize(np_img, (image_size, image_size))
    np_img = per_image_standardization(np_img)
    np_img = np_img[:, :, [2, 1, 0]]  # BGR2RGB
    return np_img.ravel()


def softmax(x):
    x_row_max = x.max(axis=-1)
    x_row_max = x_row_max.reshape(list(x.shape)[:-1] + [1])
    x = x - x_row_max
    x_exp = np.exp(x)
    x_exp_row_sum = x_exp.sum(axis=-1).reshape(list(x.shape)[:-1] + [1])
    softmax = x_exp / x_exp_row_sum
    return softmax


def convert_to_list(dict_val):
    list_key = list(dict_val)
    list_val = [i[0] for i in list(dict_val.values())]
    list_result = list(chain.from_iterable([zip(list_key, list_val)]))
    return list_result


def output_to_csv(filename, first_row, other_rows):
    with open(filename, 'w') as f:
        csv_writer = csv.writer(f)
        csv_writer.writerow(first_row)
        for row in other_rows:
            csv_writer.writerow(row)


def load_labels(label_file):
    labels_dict = {}
    if os.path.isfile(label_file):
        with open(label_file) as f:
            for line in f:
                key, value = line.rstrip('\n').split(':')
                labels_dict[key] = value
    return labels_dict


def do_inference(context, bindings, inputs, outputs, stream):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async(batch_size=1, bindings=bindings,
                          stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding))
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


# Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


class TRTInference(object):
    def load_engine(self, plan_file, label_file, model_name):
        self.labels_dict = load_labels(label_file)
        self.image_size = get_image_size(model_name)

        with open(plan_file, "rb") as f,\
                trt.Runtime(TRT_LOGGER) as runtime:
            self.engine = runtime.deserialize_cuda_engine(f.read())

    def _get_images_list(self):
        image_dict_list = []
        if os.path.isdir(self.image_path):
            self.single_infer = False
            for filepath, dirnames, filenames in os.walk(self.image_path):
                for filename in filenames:
                    image_dict_list.append({
                        "file_path": os.path.join(filepath, filename)
                    })
        else:
            self.single_infer = True
            image_dict_list.append({
                "file_path": self.image_path
            })

        return image_dict_list

    def parse_result(self, image_dict_list, start_idx, batch_output_list):
        predict_res = {}
        for idx, item in enumerate(batch_output_list):
            try:
                img_file_path = image_dict_list[start_idx + idx]['file_path']
                if self.single_infer:
                    relative_f = self.image_path
                else:
                    relative_f = img_file_path.replace(
                        self.image_path.rstrip('/'),
                        '.'
                    )

                if len(item) >= 5:
                    len_num = 5
                else:
                    len_num = len(item)
                result_list = []
                for i, j in enumerate(item):
                    result = [
                        self.labels_dict[str(i)],     # label
                        round(float(softmax(item)[i]), 4)    # probability
                    ]
                    result_list.append(result)
                predict_res[relative_f] = sorted(
                    result_list, key=lambda x: x[1], reverse=True)[0:len_num]

            except IndexError:
                break
        return predict_res

    def inference_by_engine(self, engine):
        predict_res = {}
        batch_size = engine.max_batch_size
        img_arr_len = self.image_size * self.image_size * 3
        dest_input_len = img_arr_len * batch_size

        start_time = time.time()
        image_dict_list = self._get_images_list()
        image_count = len(image_dict_list)

        with engine.create_execution_context() as context:
            inputs, outputs, bindings, stream = allocate_buffers(engine)

            for start_idx in range(0, image_count, batch_size):
                if start_idx + batch_size > image_count:
                    real_batch_size = image_count - start_idx
                else:
                    real_batch_size = batch_size

                source = load_image(
                    image_dict_list[start_idx]['file_path'], self.image_size
                )
                if real_batch_size > 1:
                    for i in range(start_idx + 1, start_idx + real_batch_size):
                        source = np.concatenate((
                            source,
                            load_image(image_dict_list[i]['file_path'],
                                       self.image_size)
                        ))
                if len(source) < dest_input_len:
                    source = np.concatenate((
                        source, np.zeros(
                            img_arr_len * (batch_size - real_batch_size)
                        )
                    ))
                np.copyto(inputs[0].host, source)
                [output] = do_inference(
                    context,
                    bindings=bindings,
                    inputs=inputs,
                    outputs=outputs,
                    stream=stream
                )
                predict_res.update(self.parse_result(
                    image_dict_list=image_dict_list,
                    start_idx=start_idx,
                    batch_output_list=np.array_split(output, batch_size)
                ))
        end_time = time.time()

        return predict_res, int(round(1000 * (end_time - start_time)))

    def inference(self, image_path):
        self.image_path = image_path
        predict_res, duration = self.inference_by_engine(self.engine)
        print("[INFO] inference duration time(ms): ", duration)

        if self.single_infer:
            ret_list = list(predict_res.values())[0]
            res = {
                "filename": list(predict_res)[0],
                "data": [{
                    "label": ret[0] if not np.isnan(ret[1]) else '-',
                    "probability": ret[1] if not np.isnan(ret[1]) else 0
                } for ret in ret_list],
                "duration": duration
            }
            print("Inference result: ", res)
            return res
        else:
            return predict_res

    def create_output_file(self, predict_res, output_path):
        if self.single_infer:
            print("[INFO] single image file inference will not "
                  "create output_file")
            return

        out_file = 'predict_result.csv'
        output_to_csv(
            os.path.join(output_path, out_file),
            ['image_path', 'predict_lable'],
            convert_to_list(predict_res)
        )
        print("[INFO] Finished. You can check inference result file: {0}"
              .format(os.path.join(os.path.abspath(output_path), out_file)))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="python sample.py")
    parser.add_argument("-i", "--image_path", required=True,
                        help="Input image path to do inference.")
    parser.add_argument("-p", "--plan_file",
                        help="TensorRT plan file, default search from the "
                             "current dir.")
    parser.add_argument("-l", "--label_file",
                        help="For inferencing to real label, default search "
                             "from the current dir.")
    parser.add_argument("-o", "--output_path", default='.',
                        help="Output directory")
    args = parser.parse_args()

    print(f"[INFO] Params: \n"
          f"image_path: {args.image_path} \n"
          f"plan_file: {args.plan_file} \n"
          f"label_file: {args.label_file} \n"
          f"output_path: {args.output_path}")

    # search plan_file & label_file from the current dir if not input
    current_dir = os.path.abspath(os.path.dirname(__file__))
    model_define_file = os.path.join(current_dir, 'model_define.json')
    if not os.path.isfile(model_define_file):
        raise FileNotFoundError(f'{model_define_file} was not found.')
    with open(model_define_file, 'r') as f:
        model_json = json.loads(f.read())
    model_name = model_json.get('model_name')

    if args.label_file:
        label_file = args.label_file
    else:
        label_file = os.path.join(current_dir, 'labels.txt')
        if not os.path.isfile(label_file):
            raise FileNotFoundError(f'{label_file} was not found.')
    if args.plan_file:
        plan_file = args.plan_file
    else:
        plan_file = os.path.join(current_dir, f'{model_name}.engine')
        if not os.path.isfile(plan_file):
            raise FileNotFoundError(f'{plan_file} was not found.')

    # inference
    inference_service_mgt = TRTInference()
    inference_service_mgt.load_engine(plan_file, label_file, model_name)
    inf_result = inference_service_mgt.inference(args.image_path)
    inference_service_mgt.create_output_file(inf_result, args.output_path)
